from collections.abc import Iterable
import kliff
from ..dataset import Configuration
from ..utils import length_equal

logger = kliff.logger.get_logger(__name__)


class Calculator:
    r"""Calculator class to exchange information between model and an optimizer.

    It computes the `energy`, `forces`, etc. using a potential model, and provides these
    properties, together with the corresponding reference data, to
    :class:`~kliff.loss.Loss` to construct a cost function for the optimizer.  In the
    reverse direction, it grab the new parameters from the optimizer and update the model
    with the new parameters.

    Parameters
    ----------
    model: obj
        An instance of the :class:`~kliff.models.Model` class.
    """

    def __init__(self, model):
        self.model = model

    def create(self, configs, use_energy=True, use_forces=True, use_stress=False):
        r"""Create compute arguments for a collection of configurations.

        By compute arguments, we mean the information needed by a model to carry on a
        calculation, such as the coordinates, species, cutoff distance, neighbor list,
        etc. Each configuration has its own compute arguments, and this function creates
        the compute arguments for all the configurations in ``configs``.

        Parameters
        ----------
        configs: list
            A list of :class:`~kliff.dataset.Configuration` instances.

        use_energy: list of bools (optional)
            Whether to require the calculator to compute energy. Each component is for one
            configuration in ``configs``. If a bool instead of a list is provided, it is
            applied to all configurations.

        use_forces: list of bools (optional)
            Whether to require the calculator to compute forces. Each component is for one
            configuration in ``configs``. If a bool instead of a list is provided, it is
            applied to all configurations.

        use_stress: list of bools (optional)
            Whether to require the calculator to compute stress. Each component is for one
            configuration in ``configs``. If a bool instead of a list is provided, it is
            applied to all configurations.
        """

        self.use_energy = use_energy
        self.use_forces = use_forces
        self.use_stress = use_stress

        if isinstance(configs, Configuration):
            configs = [configs]

        if not length_equal(configs, use_energy):
            raise InputError('Lengths of arguments "configs" and "use_energy" not equal.')
        if not length_equal(configs, use_forces):
            raise InputError('Lengths of arguments "configs" and "use_forces" not equal.')
        if not length_equal(configs, use_stress):
            raise InputError('Lengths of arguments "configs" and "use_stress" not equal.')

        N = len(configs)
        if not isinstance(use_energy, Iterable):
            use_energy = [use_energy for _ in range(N)]
        if not isinstance(use_forces, Iterable):
            use_forces = [use_forces for _ in range(N)]
        if not isinstance(use_stress, Iterable):
            use_stress = [use_stress for _ in range(N)]

        # update parameters from fitting parameters to model parameters
        # influence distance may be updated if the user set the related parameters
        self.model.update_model_params()

        supported_species = self.model.get_supported_species()
        infl_dist = self.model.get_influence_distance()

        self.compute_arguments = []
        for conf, e, f, s in zip(configs, use_energy, use_forces, use_stress):
            if self._is_kim_model():
                kim_ca = self.model.create_a_kim_compute_argument()
                ca = self.model.compute_arguments_class(
                    kim_ca, conf, supported_species, infl_dist, e, f, s
                )
            else:
                ca = self.model.compute_arguments_class(
                    conf, supported_species, infl_dist, e, f, s
                )
            self.compute_arguments.append(ca)

        logger.info('calculator for %d configurations created.', len(configs))
        return self.compute_arguments

    def _is_kim_model(self):
        return self.model.__class__.__name__ == 'KIM'

    def get_compute_arguments(self):
        r"""Return a list of compute arguments, each associated with a configuration."""
        return self.compute_arguments

    # TODO, maybe change the compute_argument.compute api of kim models, such
    # that it accept params as the argument, insteand of kim_model
    def compute(self, compute_arguments):
        r"""Compute the properties given the compute arguments associated with a
        configuration.

        Parameters
        ----------
        compute_arguments: obj
            A compute arguments instance for a configuration.

        Return
        ------
        dict
            A dictionary of properties, with keys of `energy`, `forces` and `stress`.
        """
        if self._is_kim_model():
            compute_arguments.compute(self.model.kim_model)
        else:
            compute_arguments.compute(self.model.params)
        return compute_arguments.results

    # TODO, possibly, and an argument `reference` to get reference values
    def get_energy(self, compute_arguments):
        r"""Get the energy of a configuration.

        Parameters
        ----------
        compute_arguments: obj
            A compute arguments instance for a configuration.

        Return
        ------
        float
            The energy of the configuration associated with the compute arguments.
        """
        return compute_arguments.get_energy()

    def get_forces(self, compute_arguments):
        r"""Get the forces of a configuration.

        Parameters
        ----------
        compute_arguments: obj
            A compute arguments instance for a configuration.

        Return
        ------
        2D array
            The forces of the configuration associated with the compute arguments.
        """

        return compute_arguments.get_forces()

    def get_stress(self, compute_arguments):
        r"""Get the stress of a configuration.

        Parameters
        ----------
        compute_arguments: obj
            A compute arguments instance for a configuration.

        Return
        ------
        1D array
            The stress of the configuration associated with the compute arguments. The
            returned stress is in Voigt notation, i.e. this function returns:
            [:math:`\sigma_{xx}, \sigma_{yy}, \sigma_{zz}, \sigma_{yz}, \sigma_{xy},
            \sigma_{xz}`]
        """

        return compute_arguments.get_stress()

    # TODO maybe move this and `get_reference` to loss.
    def get_prediction(self, compute_arguments):
        r"""Get the prediction of all properties that are requested to compute.

        The `energy`, `forces`, and `stress` are each flattened to a 1D array, and then
        concatenated (in the order of `energy`, `forces`, and `stress`) to form the
        prediction.  Depending on the values of ``use_energy``, ``use_forces``, and
        ``use_stress`` that are provided in :meth:`create`, one or more of `energy`,
        `forces` and `stress` may not be included in the prediction.

        Parameters
        ----------
        compute_arguments: obj
            A compute arguments instance for a configuration.

        Return
        ------
        1D array
            For a configuration of N atoms:

            - If ``use_energy``, ``use_forces``, and ``use_stress`` are all ``True``, the
              size of prediction is `1+3N+6`, with the 1st component the `energy`, the 2nd
              to 1+3N components the flattened `forces` on N atoms, and the remaining 6
              components the Voigt `stress`.
            - If one or more of ``use_energy``, ``use_forces``, and ``use_stress`` is
              ``False``, its corresponding value is removed from prediction, and the size
              of prediction shrinks accordingly. For example, if both ``use_energy`` and
              ``use_stress`` are ``True`` but ``use_forces`` is ``False``, then the size
              of prediction is `1+6`, with the 1st component the `energy`, and the 2nd to
              7th components the Voigt stress.
        """
        return compute_arguments.get_prediction()

    def get_reference(self, compute_arguments):
        r"""Get the reference data of all properties that are requested to compute.

        Same as :meth:`get_prediction`, the `energy`, `forces`, and `stress` are each
        flattened to a 1D array, and then concatenated (in the order of `energy`,
        `forces`, and `stress`) to form the reference. Depending on the values of
        ``use_energy``, ``use_forces``, and ``use_stress`` that are provided in
        :meth:`create`, one or more of `energy`, `forces` and `stress` may not be included
        in the reference.

        Parameters
        ----------
        compute_arguments: obj
            A compute arguments instance for a configuration.

        Return
        ------
        1D array
            For a configuration of N atoms:

            - If ``use_energy``, ``use_forces``, and ``use_stress`` are all ``True``, the
              size of reference is `1+3N+6`, with the 1st component the `energy`, the 2nd
              to 1+3N components the flattened `forces` on N atoms, and the remaining 6
              components the Voigt `stress`.
            - If one or more of ``use_energy``, ``use_forces``, and ``use_stress`` is
              ``False``, its corresponding value is removed from reference, and the size
              of reference shrinks accordingly. For example, if both ``use_energy`` and
              ``use_stress`` are ``True`` but ``use_forces`` is ``False``, then the size
              of reference is `1+6`, with the 1st component the `energy`, and the 2nd to
              7th components the Voigt stress.
        """
        return compute_arguments.get_reference()

    def get_opt_params(self):
        r"""Return a list of optimizing parameters.

        The optimizing parameters is a list consisting of the values of the model
        parameters that is set to fit via :meth:`kliff.models.Model.set_fitting_params` or
        :meth:`kliff.models.Model.set_one_fitting_param`.  The returned value is typically
        passed to an optimizer as the initial values to carry out the optimization.
        """
        return self.model.get_opt_params()

    def has_opt_params_bounds(self):
        r"""Return a bool to indicate whether there are parameters whose bounds are
        provided.
        """
        return self.model.has_opt_params_bounds()

    def get_opt_params_bounds(self):
        r"""Return the lower and upper bounds for the optimizing parameters.

        The returned value is a list of (lower, upper) tuples. Each tuple contains the
        lower and upper bounds for the corresponding parameter obtained from
        :meth:`get_opt_params`. ``None`` for ``lower`` or ``upper`` means that no bound
        should be applied.
        """
        return self.model.get_opt_params_bounds()

    def update_opt_params(self, opt_params):
        r"""Update the optimizing parameters from optimizer to model.

        This function is the reverse of :meth:`get_opt_params`.

        Parameters
        ----------
        opt_params: list
            Optimizing parameters.
        """
        self.model.update_fitting_params(opt_params)
        self.model.apply_params_relation()
        self.model.update_model_params()


class _WrapperCalculator(object):
    """Wrapper to deal with the fitting of multiple models."""

    def __init__(self, *calculators):
        """
        Parameters
        ----------

        calculators: instance of Calculator
        """
        self.calculators = calculators
        self._start_end = self._set_start_end()

    def _set_start_end(self):
        """Compute the start and end indices of the `opt_params` of each calculator in the
        `opt_params` of the wrapper calculator."""
        start_end = []
        i = 0
        for calc in self.calculators:
            N = calc.get_number_of_opt_params()
            start = i
            end = i + N
            start_end.append((start, end))
            i += N
        return start_end

    def get_compute_arguments(self):
        all_cas = []
        for calc in self.calculators:
            cas = calc.get_compute_arguments()
            all_cas.extend(cas)
        return all_cas

    def get_number_of_opt_params(self):
        N = 0
        for calc in self.calculators:
            N += calc.get_number_of_opt_params()
        return N

    def get_opt_params(self):
        opt_params = []
        for calc in self.calculators:
            p = calc.get_opt_params()
            opt_params.extend(p)
        return opt_params

    def get_opt_params_bounds(self):
        bounds = []
        for calc in self.calculators:
            b = calc.get_opt_params_bounds()
            bounds.extend(b)
        return bounds

    def update_model_params(self):
        for calc in self.calculators:
            calc.update_model_params()

    def update_params(self, opt_params):
        for i, calc in enumerate(self.calculators):
            start, end = self._start_end[i]
            p = opt_params[start:end]
            calc.update_params(p)

    def get_calculator_list(self):
        """Create a list of calculators.

        Each calculator has `number of configurations` copies in the list.
        """
        calc_list = []
        for calc in self.calculators:
            N = len(calc.get_compute_arguments())
            calc_list.extend([calc] * N)
        return calc_list



import numpy as np
import os
import pytest
from kliff.calculators import Calculator
from kliff.models import LennardJones
from kliff.models.model import ModelError
from kliff.dataset import Dataset


def params_relation(params):
    sigma = params.get_value('sigma')
    epsilon = params.get_value('epsilon')
    epsilon[0] = 2 * sigma[0]
    params.set_value('epsilon', epsilon)


def write_tmp_params(fname):
    with open(fname, 'w') as fout:
        fout.write('sigma\n')
        fout.write('1.1  fix\n')
        fout.write('epsilon\n')
        fout.write('2.1  None  3.\n')


def delete_tmp_params(fname):
    os.remove(fname)


def energy_forces_stress(
    calc, configs, use_energy=False, use_forces=False, use_stress=False
):

    pred_energy = -56.08345486063805
    pred_forces = [
        [2.41107532e-02, 1.29215347e-03, 2.89182619e-04],
        [-2.13103656e-02, -7.23000337e-03, -1.28954011e-02],
        [3.89036609e-04, -2.07124189e-03, -3.83185414e-01],
        [7.03144295e-04, 4.09004596e-04, -3.62012614e-01],
        [-1.84061312e-03, 6.56687434e-03, 3.62871504e-01],
        [-6.79944094e-03, 6.50195868e-03, 3.95977434e-01],
    ]
    pred_stress = [
        4.24791582e-03,
        4.26804209e-03,
        4.41900170e-03,
        -3.25993799e-06,
        -2.05334688e-06,
        -2.82003088e-06,
    ]

    ref_energy = -5.302666
    ref_forces = [
        [-0.425324, 0.295866, -0.065479],
        [0.245043, -0.061658, 0.264104],
        [0.010127, 0.041539, 0.301571],
        [0.079468, -0.072558, -0.340646],
        [0.126296, -0.152711, 0.313083],
        [0.224514, -0.233065, -0.737724],
    ]
    ref_stress = [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

    calc.create(configs, use_energy, use_forces, use_stress)
    cas = calc.get_compute_arguments()
    ca = cas[0]
    calc.compute(ca)

    assert ca.get_compute_flag('energy') == use_energy
    assert ca.get_compute_flag('forces') == use_forces
    assert ca.get_compute_flag('stress') == use_stress

    try:
        energy = calc.get_energy(ca)
        assert energy == pytest.approx(pred_energy, 1e-6)
    except ModelError as e:
        if use_energy:
            raise ModelError(e)
        else:
            pass

    try:
        forces = calc.get_forces(ca)
        assert np.allclose(forces[:6], pred_forces)
    except ModelError as e:
        if use_forces:
            raise ModelError(e)
        else:
            pass

    try:
        stress = calc.get_stress(ca)
        assert np.allclose(stress, pred_stress)
    except ModelError as e:
        if use_stress:
            raise ModelError(e)
        else:
            pass

    pred = calc.get_prediction(ca)
    ref = calc.get_reference(ca)

    if use_energy:
        assert pred[0] == pytest.approx(pred_energy, 1e-6)
        assert ref[0] == pytest.approx(ref_energy, 1e-6)
    if use_forces:
        if use_energy:
            assert np.allclose(pred[1 : 1 + 3 * 6], np.ravel(pred_forces))
            assert np.allclose(ref[1 : 1 + 3 * 6], np.ravel(ref_forces))
        else:
            assert np.allclose(pred[: 3 * 6], np.ravel(pred_forces))
            assert np.allclose(ref[: 3 * 6], np.ravel(ref_forces))

    if use_stress:
        assert np.allclose(pred[-6:], pred_stress)
        assert np.allclose(ref[-6:], ref_stress)


def test_lj():
    model = LennardJones()

    # set params directly
    model.set_fitting_params(sigma=[[1.1, 'fix']], epsilon=[[2.1, None, 3.0]])

    model.update_model_params()
    # model.echo_model_params()
    # model.echo_fitting_params()

    # set params by reading from file (the same as set params directly)
    # fname = 'tmp_lj.params'
    # write_tmp_params(fname)
    # model.read_fitting_params(fname)
    # delete_tmp_params(fname)

    calc = Calculator(model)

    dset = Dataset(order_by_species=False)
    fname = './configs_extxyz/MoS2/MoS2_energy_forces_stress.xyz'
    dset.read(fname)
    configs = dset.get_configs()

    energy_forces_stress(calc, configs, True, False, False)
    energy_forces_stress(calc, configs, True, True, False)
    energy_forces_stress(calc, configs, True, True, True)

    # params relation callback
    model.set_params_relation_callback(params_relation)
    x0 = calc.get_opt_params()
    calc.update_opt_params(x0)
    sigma = model.get_model_params('sigma')
    epsilon = model.get_model_params('epsilon')
    assert np.allclose(sigma * 2, epsilon)


if __name__ == '__main__':
    test_lj()



import kliff
from kliff.dataset import Dataset
from kliff.calculator import Calculator
from kliff.models import LennardJones
from kliff.loss import Loss

kliff.logger.set_level('debug')

# training set
tset = Dataset()
tset.read('../tests/configs_extxyz/Si_4/')
configs = tset.get_configs()

# calculator
model = LennardJones()
# model.echo_model_params()

# fitting parameters
model.set_fitting_params(sigma=[['default']], epsilon=[['default']])
# model.echo_fitting_params()

calc = Calculator(model)
calc.create(configs)

# loss
with Loss(calc, nprocs=1) as loss:
    result = loss.minimize(method='L-BFGS-B', options={'disp': True, 'maxiter': 10})


# print optimized parameters
model.echo_fitting_params()
model.save_model_params()



"""Create a separate page for each module.

To run:
    $ python genereate_apidoc.py
which generates a directory named `tmp_apidoc`. Do a comparison of this file

You can provide an `exclude` list.

"""
import os
import subprocess


def get_all_modules(source='../kliff'):
    """Get all modules.
    Note, this only get the first-level modules like `kliff.module_a`, not modules
    (in subpackages) like `kliff.subpackage_a.module_b`. subpackage is considered
    as a module.

    Take advantage of
        $ sphinx-apidoc -f -e -o <outdir> <sourcedir>
    Return a list of modules names.
    """
    results = subprocess.check_output(
        ['sphinx-apidoc', '-f', '-e', '-o', '/tmp/kliff_apidoc', source],
        universal_newlines=True,
    )
    results = results.split('\n')
    modules = []
    for line in results:
        if 'Creating' in line:
            name = line.rstrip('.').split('/')[-1].split('.')
            if len(name) >= 3:
                mod = name[1]
                if mod not in modules:
                    modules.append(mod)
    return modules


def autodoc_package(path, modules):
    if path and not os.path.exists(path):
        os.makedirs(path)
    fname = os.path.join(path, 'kliff.rst')
    with open(fname, 'w') as fout:
        fout.write('.. _reference:\n\n')
        fout.write('Package Reference\n')
        fout.write('=================\n\n')
        fout.write('.. toctree::\n')
        for mod in modules:
            fout.write('    kliff.' + mod + '\n')


def autodoc_module(path, module):
    if path and not os.path.exists(path):
        os.makedirs(path)
    module_name = 'kliff.' + module
    fname = os.path.join(path, module_name + '.rst')
    with open(fname, 'w') as fout:
        fout.write('{}\n'.format(module_name))
        fout.write('-' * len(module_name) + '\n\n')
        fout.write('.. automodule:: {}\n'.format(module_name))
        fout.write('    :members:\n')
        fout.write('    :undoc-members:\n')
        fout.write('    :show-inheritance:\n')
        fout.write('    :inherited-members:\n')


if __name__ == '__main__':

    # this will not be included, e.g. excludes = ['models']
    excludes = ['cmdline', 'lmplatconst', 'wrapcalculator']

    modules = get_all_modules()
    for exc in excludes:
        modules.remove(exc)
    modules = sorted(modules)
    path = './tmp_apidoc'
    autodoc_package(path, modules)
    for mod in modules:
        autodoc_module(path, mod)



from setuptools import setup, Extension, find_packages
from distutils.sysconfig import get_config_vars
import os


# remove `-Wstrict-prototypes' that is for C not C++
cfg_vars = get_config_vars()
for key, value in cfg_vars.items():
    if type(value) == str and '-Wstrict-prototypes' in value:
        cfg_vars[key] = value.replace('-Wstrict-prototypes', '')


class get_pybind11_includes:
    """Helper class to determine the pybind11 include path.

    The purpose of this class is to postpone importing pybind11 until it is actually
    installed, so that the ``get_include()`` method can be invoked.

    See:
    https://github.com/pybind/python_example/blob/master/setup.py
    https://github.com/pybind/python_example/issues/32
    """

    def __init__(self, user=False):
        try:
            import pybind11
        except ImportError:
            if subprocess.call([sys.executable, '-m', 'pip', 'install', 'pybind11']):
                raise RuntimeError('pybind11 install failed.')
        self.user = user

    def __str__(self):
        import pybind11

        return pybind11.get_include(self.user)


def get_includes():
    return [get_pybind11_includes(), get_pybind11_includes(user=True)]


def get_extra_compile_args():
    return ['-std=c++11']


sym_fn = Extension(
    'kliff.descriptors.symmetry_function.sf',
    sources=[
        'kliff/descriptors/symmetry_function/sym_fn_bind.cpp',
        'kliff/descriptors/symmetry_function/sym_fn.cpp',
        'kliff/descriptors/symmetry_function/helper.cpp',
    ],
    include_dirs=get_includes(),
    extra_compile_args=get_extra_compile_args(),
    language='c++',
)

bispectrum = Extension(
    'kliff.descriptors.bispectrum.bs',
    sources=[
        'kliff/descriptors/bispectrum/bispectrum_bind.cpp',
        'kliff/descriptors/bispectrum/bispectrum.cpp',
        'kliff/descriptors/bispectrum/helper.cpp',
    ],
    include_dirs=get_includes(),
    extra_compile_args=get_extra_compile_args(),
    language='c++',
)


def get_version(fname=os.path.join('kliff', '__init__.py')):
    with open(fname) as fin:
        for line in fin:
            line = line.strip()
            if '__version__' in line:
                v = line.split('=')[1]
                # stripe white space, and ' or " in string
                if "'" in v:
                    version = v.strip("' ")
                elif '"' in v:
                    version = v.strip('" ')
                break
    return version


kliff_scripts = ['bin/kliff']


setup(
    name='kliff',
    version=get_version(),
    # NOTE, subpackages need to be specified as well
    # packages=['kliff'],
    # NOTE, subpackages need to be included as well
    # packages=['kliff','tensorflow_op', 'geolm'],
    # package_dir={'geolm':'libs/geodesicLMv1.1/pythonInterface'},
    # package_data={'geolm':['_geodesiclm.so']},
    packages=find_packages(),
    ext_modules=[sym_fn, bispectrum],
    scripts=kliff_scripts,
    install_requires=['scipy', 'pybind11', 'pytest'],
    author='Mingjian Wen',
    author_email='wenxx151@gmail.com',
    url='https://github.com/mjwen/kliff',
    description='KLIFF: KIM-based Learning-Integrated Fitting Framework',
    long_description='KLIFF: KIM-based Learning-Integrated Fitting Framework',
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: Common Development and Distribution License 1.0 (CDDL-1.0)',
        'Operating System :: OS Independent',
    ],
    zip_safe=False,
)



import numpy as np
import warnings
from collections import OrderedDict
from kliff.models.parameter import Parameter
from kliff.models.parameter import FittingParameter
from kliff.models.parameter import ParameterError


def test_parameter():

    # scalar
    try:
        p = Parameter(1.1)
    except ParameterError:
        pass

    # 2D array
    try:
        p = Parameter([[1.1]])
    except ParameterError:
        pass

    p = Parameter([2.2, 3.3])
    assert np.allclose(p.get_value(), [2.2, 3.3])
    assert p.get_size() == 2
    assert p.get_dtype() == 'double'
    assert p.get_description() == None

    p.set_value([4.4])
    assert np.allclose(p.get_value(), [4.4])


def create_all_possible_input(v):
    inp = []
    for ud in ['default', v + 0.01]:
        inp.append([ud])
        inp.append([ud, 'fix'])
        for lb in [None, v - 0.1]:
            for ub in [None, v + 0.1]:
                inp.append([ud, lb, ub])
    return inp


def test_fitting_parameter():

    # model params
    mp = OrderedDict()
    p1 = [1.1]
    p2 = [2.2, 3.3]
    names = ['p1', 'p2']
    mp[names[0]] = Parameter(p1)
    mp[names[1]] = Parameter(p2)
    fp = FittingParameter(mp)

    # fitting params
    with warnings.catch_warnings():  # context manager to ignore warning
        warnings.simplefilter('ignore')

        v0 = p1[0]
        inp0 = create_all_possible_input(v0)
        v1 = p2[0]
        inp1 = create_all_possible_input(v1)
        v2 = p2[1]
        inp2 = create_all_possible_input(v2)
        for i in inp0:
            for j in inp1:
                for k in inp2:
                    fp.set(p1=[i])
                    fp.set(p2=[j, k])

    assert np.allclose(fp.get_opt_params(), [1.11, 2.21, 3.31])

    # interface to optimizer
    # restore to default parameter values
    x0 = np.concatenate((p1, p2))
    fp.update_params(x0)
    assert np.allclose(fp.get_opt_params(), x0)

    assert fp.get_number_of_opt_params() == 3

    for i in range(3):
        n, v, p, c = fp.get_opt_param_name_value_and_indices(i)
        assert v == x0[i]
        if i == 0:
            assert n == 'p1'
            assert p == 0
            assert c == 0
        else:
            assert n == 'p2'
            assert p == 1
            assert c == (i - 1) % 2

    bounds = [[i - 0.1, i + 0.1] for i in x0]
    assert np.allclose(fp.get_opt_params_bounds(), bounds)

    # interface to calculator
    nms = fp.get_names()
    for i, j in zip(nms, names):
        assert i == j

    for i, nm in enumerate(nms):
        assert fp.get_size(nm) == mp[nm].size
        assert np.allclose(fp.get_value(nm), mp[nm].value)
        assert np.allclose(fp.get_lower_bound(nm), [v - 0.1 for v in mp[nm].value])
        assert np.allclose(fp.get_upper_bound(nm), [v + 0.1 for v in mp[nm].value])
        fix = fp.get_fix(nm)
        for x in fix:
            assert not x


if __name__ == '__main__':
    test_parameter()
    test_fitting_parameter()



from .dataset import Configuration
from .dataset import Dataset
from .dataset import read_config
from .dataset import write_config

__all__ = ['Configuration', 'Dataset', 'read_config', 'write_config']



import numpy as np
from kliff.dataset import Configuration
from kliff.dataset import Dataset
from kliff.dataset.extxyz import write_extxyz


def assert_1d_array(a, b):
    for i, j in zip(a, b):
        assert i == j


def test_configuration(e=True, f=False, s=False, order=False):
    fname = './configs_extxyz/MoS2/MoS2'
    if e:
        fname += '_energy'
    if f:
        fname += '_forces'
    if s:
        fname += '_stress'
    fname += '.xyz'

    config = Configuration(format='extxyz', order_by_species=order)
    config.read(fname)

    natoms = config.get_number_of_atoms()
    cell = config.get_cell()
    PBC = config.get_PBC()
    species = config.get_species()
    coords = config.get_coordinates()
    if e:
        energy = config.get_energy()
    if f:
        forces = config.get_forces()
    if s:
        stress = config.get_stress()

    assert natoms == 288
    assert np.allclose(
        cell,
        [
            [33.337151, 0.035285, 0.03087],
            [0.027151, 25.621674, -0.000664],
            [0.027093, 9e-05, 30.262626],
        ],
    )
    assert np.allclose(PBC, [1, 1, 1])

    if e:
        assert energy == -5.302666

    if s:
        assert np.allclose(stress, [1.1, 2.2, 3.3, 4.4, 5.5, 6.6])

    if order:
        ref_species = ['Mo', 'Mo', 'Mo', 'Mo', 'Mo', 'Mo']
        ref_coords = [
            [0.051823, -0.017150, 16.736001],
            [2.819495, 1.607633, 16.734550],
            [5.601008, 0.015314, 16.738646],
        ]
        ref_forces = [
            [-0.425324, 0.295866, -0.065479],
            [0.245043, -0.061658, 0.264104],
            [-0.164382, -0.103679, -0.050262],
        ]
    else:
        ref_species = ['Mo', 'Mo', 'S', 'S', 'S', 'S']
        ref_coords = [
            [0.051823, -0.017150, 16.736001],
            [2.819495, 1.607633, 16.734550],
            [1.901722, -0.004881, 18.317009],
        ]
        ref_forces = [
            [-0.425324, 0.295866, -0.065479],
            [0.245043, -0.061658, 0.264104],
            [0.010127, 0.041539, 0.301571],
        ]

    assert_1d_array(species[:6], ref_species)
    assert np.allclose(coords[:3], ref_coords)
    if f:
        assert np.allclose(forces[:3], ref_forces)

    natoms_by_species = config.count_atoms_by_species()
    assert natoms_by_species['Mo'] == 96
    assert natoms_by_species['S'] == 192

    if not e:
        energy = None
    if not f:
        forces = None
    if not s:
        stress = None
    outname = fname.replace('./configs_extxyz/MoS2/MoS2', 'tmp_Mos2')
    write_extxyz(outname, cell, PBC, species, coords, energy, forces, stress)


def test_dataset():
    directory = './configs_extxyz/MoS2'
    tset = Dataset()
    tset.read(directory)
    configs = tset.get_configs()
    assert len(configs) == 3


if __name__ == '__main__':
    test_configuration(True, False, False)
    test_configuration(True, True, False)
    test_configuration(True, True, True)
    test_configuration(True, True, True, order=True)
    test_dataset()



import os
import numpy as np
from collections import OrderedDict
import kliff
from .extxyz import read_extxyz, write_extxyz
from ..log import log_entry

logger = kliff.logger.get_logger(__name__)


implemented_format = dict()
implemented_format['extxyz'] = '.xyz'


class Configuration:
    r"""Class of atomic configuration.

    Parameters
    ----------
    format: str
        Format of the file that stores the configuration. Currently, supported format
        includes: `extxyz`.

    identifier: str
        A unique identifier of the configuration.

    order_by_species: bool
        If `True`, the atoms in the configuration will be ordered according to their
        species such that atoms with the same species will have contiguous indices.
    """

    def __init__(self, format='extxyz', identifier=None, order_by_species=True):
        self.format = format
        self.id = identifier
        self.do_order = order_by_species
        self.weight = 1.0
        self.natoms = None  # int
        self.cell = None  # ndarray of shape(3,3)
        self.PBC = None  # ndarray of shape(3,)
        self.energy = None  # float
        self.stress = None  # ndarray of shape(6,)
        self.species = None  # ndarray of shape(N,)
        self.coords = None  # ndarray of shape(N, 3)
        self.forces = None  # ndarray of shape(N, 3)
        self.natoms_by_species = None  # dict

    def read(self, path):
        r"""Read configuration stored in a file.

        Parameters
        ----------
        path: str
            Path to the file that stores the configuration.
        """
        (
            self.cell,
            self.PBC,
            self.species,
            self.coords,
            self.energy,
            self.forces,
            self.stress,
        ) = read_config(path, self.format)
        self.natoms = len(self.species)
        self.volume = abs(np.dot(np.cross(self.cell[0], self.cell[1]), self.cell[2]))

        if self.do_order:
            self.order_by_species()

    def order_by_species(self):
        r"""Order the atoms according to the species."""
        if self.forces is not None:
            species, coords, forces = zip(
                *sorted(
                    zip(self.species, self.coords, self.forces), key=lambda pair: pair[0]
                )
            )
            self.species = np.asarray(species)
            self.coords = np.asarray(coords)
            self.forces = np.asarray(forces)
        else:
            species, coords = zip(
                *sorted(zip(self.species, self.coords), key=lambda pair: pair[0])
            )
            self.species = np.asarray(species)
            self.coords = np.asarray(coords)

    def count_atoms_by_species(self, symbols=None):
        r"""Count the number of atoms with species `symbols` in the configuration.

        Parameters
        ----------
        symbols: list
            A list of species string. If `None`, the species that are already in the
            configuration are used.

        Return
        ------
        Return an OrderedDict with `keys` the species string specified in `symbols`, and
        `values` the number of atoms with each species.
        """

        unique, counts = np.unique(self.species, return_counts=True)
        symbols = unique if symbols is None else symbols

        natoms_by_species = OrderedDict()
        for s in symbols:
            if s in unique:
                natoms_by_species[s] = counts[list(unique).index(s)]
            else:
                natoms_by_species[s] = 0

        self.natoms_by_species = natoms_by_species
        return natoms_by_species

    def get_identifier(self):
        r"""Return the identifier of the configuration, which is specified at the
        initialization of the class."""
        return self.id

    def get_number_of_atoms(self):
        r"""Return the total number of atoms in the configuration."""
        return self.natoms

    def get_number_of_atoms_by_species(self):
        r"""Return a dictionary of the number of atoms with each species."""
        return self.count_atoms_by_species()

    def get_cell(self):
        r"""Return a 3x3 matrix of the lattice vectors of the configurations.

        The first, second, and third rows are :math:`a_1`, :math:`a_2`, and :math:`a_3`,
        respectively.
        """
        return self.cell.copy()

    def get_volume(self):
        r"""Return the volume of the configuration."""
        return self.volume

    def get_PBC(self):
        r"""Return a list with 3 components indicating whether periodic boundary condition
        is used along the directions of the first, second, and third lattice vectors.
        """
        return self.PBC.copy()

    def get_species(self):
        r"""Return a list of species string of all atoms."""
        return self.species.copy()

    def get_coordinates(self):
        r"""Return a `Nx3` matrix of the Cartesian coordinates of all atoms."""
        return self.coords.copy()

    def get_energy(self):
        r"""Return the potential energy of the configuration."""
        if self.energy is None:
            raise DatasetError('Configuration does not contain forces.')
        return self.energy

    def get_forces(self):
        r"""Return a `Nx3` matrix of the forces on each atoms."""
        if self.forces is None:
            raise DatasetError('Configuration does not contain forces.')
        return self.forces.copy()

    def get_stress(self):
        r"""Return the stress of the configuration.

        It returns a list with 6 components in Voigt notation, i.e. it returns
        :math:`\sigma=[\sigma_{xx},\sigma_{yy},\sigma_{zz},\sigma_{yz},\sigma_{xz},
        \sigma_{xy}]`.

        See Also
        --------
            https://en.wikipedia.org/wiki/Voigt_notation
        """
        if self.stress is None:
            raise DatasetError('Configuration does not contain stress.')
        return self.stress.copy()

    def set_weight(self, weight):
        r"""Set the weight of the configuration if the loss function.

        Parameters
        ----------
        weight: float
            The weight of the configuration.
        """
        self.weight = weight

    def get_weight(self):
        r"""Get the weight of the configuration if the loss function."""
        return self.weight


class Dataset:
    r"""Dataset class to deal with multiple :class:`~kliff.dataset.Configuration`.

    Parameters
    ----------
    order_by_species: bool
        If `True`, the atoms in each configuration will be ordered according to their
        species such that atoms with the same species will have contiguous indices.

    """

    def __init__(self, order_by_species=True):
        self.do_order = order_by_species
        self.configs = []

        logger.info('"{}" instantiated.'.format(self.__class__.__name__))

    def read(self, path, fmt='extxyz'):
        r"""Read an atomic configuration.

        Parameters
        ----------
        path: str
            Path of a file storing a configuration or path to a directory containing
            multiple files. If given a directory, all the files in this directory and its
            subdirectories with the extension corresponding to the specified format will
            be read.

        fmt: str
            Format of the file that stores the configuration (e.g. 'extxyz').
        """
        try:
            extension = implemented_format[fmt]
        except KeyError as e:
            report_error('{}\nNot supported data file format "{}".'.format(e, fmt))

        if os.path.isdir(path):
            dirpath = path
            all_files = []
            for root, dirs, files in os.walk(dirpath):
                for f in files:
                    if f.endswith(extension):
                        all_files.append(os.path.join(root, f))
            all_files = sorted(all_files)
        else:
            dirpath = os.path.dirname(path)
            all_files = [path]

        configs = []
        for f in all_files:
            conf = Configuration(fmt, f, self.do_order)
            conf.read(f)
            configs.append(conf)

        size = len(configs)
        if size <= 0:
            report_error(
                'No dataset file with format "{}" found "{}".'.format(fmt, dirpath)
            )
        self.configs.extend(configs)

        if self.do_order:
            # find species present in all configurations
            all_species = []
            for conf in self.configs:
                conf_species = set(conf.get_species())
                all_species.extend(conf_species)
            all_species = set(all_species)
            # find occurrence of species in each configuration
            for conf in self.configs:
                conf.natoms_by_species = conf.count_atoms_by_species(all_species)

        msg = '{} configurations read from "{}"'.format(len(configs), path)
        log_entry(logger, msg, level='info')

    def get_configs(self):
        r"""Get the configurations.

        Return
        ------
            A list of :class:`~kliff.dataset.Configuration` instance.
        """
        return self.configs

    def get_num_configs(self):
        """Return the number of configurations in the dataset"""
        return len(self.configs)


def read_config(path, fmt='extxyz'):
    r"""Read configuration stored in a file.

    Parameters
    ----------
    path: str
        Path to the file that stores the configuration.

    fmt: str
        Format of the file that stores the configuration (e.g. `extxyz`).

    Returns
    -------
    cell: array
        A 3x3 matrix of the lattice vectors.  The first, second, and third rows are
        :math:`a_1`, :math:`a_2`, and :math:`a_3`, respectively.

    PBC: list
        A list with 3 components indicating whether periodic boundary condition is used
        along the directions of the first, second, and third lattice vectors.

    species: list
        A list of string with N component, where N is the number of atoms.

    coords: array
        A Nx3 matrix of the coordinates of the atoms, where N is the number of atoms.

    energy: float or None
        Potential energy of the configuration. If it is not provided in the file, return
        `None`.

    forces: array or None
        A Nx3 array of the forces on atoms, where N is the number of atoms. If the forces
        are not provided in the file, return `None`.

    stress: list or None
        A list with 6 components in Voigt notation, i.e. it returns
        :math:`\sigma=[\sigma_{xx},\sigma_{yy},\sigma_{zz},\sigma_{yz},\sigma_{xz},
        \sigma_{xy}]`. If the stresses are not provided in the file, return `None`.
    """

    if fmt not in implemented_format:
        report_error('Data file format "{}" not recognized.')

    if fmt == 'extxyz':
        cell, PBC, species, coords, energy, forces, stress = read_extxyz(path)

    return cell, PBC, species, coords, energy, forces, stress


def write_config(
    path, cell, PBC, species, coords, energy=None, forces=None, stress=None, fmt='extxyz'
):
    r"""
    Write a configuration to a file in the specified format.

    Parameters
    ----------
    path: str
        Path to the file that stores the configuration.

    fmt: str
        Format of the file that stores the configuration (e.g. `extxyz`).

    cell: array
        A 3x3 matrix of the lattice vectors.  The first, second, and third rows are
        :math:`a_1`, :math:`a_2`, and :math:`a_3`, respectively.

    PBC: list
        A list with 3 components indicating whether periodic boundary condition is used
        along the directions of the first, second, and third lattice vectors.

    species: list
        A list of string with N component, where N is the number of atoms.

    coords: array
        A Nx3 matrix of the coordinates of the atoms, where N is the number of atoms.

    energy: float (optional)
        Potential energy of the configuration. If `None`, skip writing this information.

    forces: array (optional)
        A Nx3 array of the forces on atoms, where N is the number of atoms. If `None`,
        skip writing this information.

    stress: list (optional)
        A list with 6 components in Voigt notation, i.e. it returns
        :math:`\sigma=[\sigma_{xx},\sigma_{yy},\sigma_{zz},\sigma_{yz},\sigma_{xz},
        \sigma_{xy}]`. If `None`, skip writing this information.
    """

    if fmt not in implemented_format:
        report_error('Data file format "{}" not recognized.')

    dirname = os.path.dirname(os.path.abspath(path))
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    if fmt == 'extxyz':
        write_extxyz(path, cell, PBC, species, coords, energy, forces, stress)


class DatasetError(Exception):
    def __init__(self, msg):
        super(DatasetError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


def report_error(msg):
    log_entry(logger, msg, level='error')
    raise DatasetError(msg)



"""
Fisher information for the SW potential.
"""


from kliff.models import KIM
from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.analyzers import Fisher


##########################################################################################
# Select the parameters that will be used to compute the Fisher information. Only
# parameters specified below will be use, others will be kept fixed. The size of the
# Fisher information matrix will be equal to the total size of the parameters specified
# here.
model = KIM(model_name='SW_StillingerWeber_1985_Si__MO_405512056662_005')
model.set_fitting_params(
    A=[['default']], B=[['default']], sigma=[['default']], gamma=[['default']]
)

# dataset
dataset_name = 'tmp_tset'
tset = Dataset()
tset.read(dataset_name)
configs = tset.get_configs()

# calculator
calc = Calculator(model)
calc.create(configs)

##########################################################################################
# Fisher information analyzer.
analyzer = Fisher(calc)
analyzer.run()
